/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/


#include "wa_def.h"

#define ADLP_GT_REV_ID_A0   SI_REV_ID(0,0)
#define ADLP_GT_REV_ID_B0   SI_REV_ID(4,4)
#define ADLP_GT_REV_ID_C0   SI_REV_ID(8,8)
