/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// Platform Revision Ids for PVC

// AUTO-GENERATED FILE. DO NOT EDIT.

#include "wa_def.h"

#define PVC_GT_REV_ID_COMPUTETILE_A0   SI_REV_ID(0,0)
#define PVC_GT_REV_ID_COMPUTETILE_A0P  SI_REV_ID(0,0)
#define PVC_GT_REV_ID_COMPUTETILE_XTA0 SI_REV_ID(3,3)
#define PVC_GT_REV_ID_COMPUTETILE_XTB0 SI_REV_ID(5,5)

#define PVC_GT_REV_ID_BASEDIE_A0   SI_REV_ID(0,0)
#define PVC_GT_REV_ID_BASEDIE_B0   SI_REV_ID(3,3)


