;=========================== begin_copyright_notice ============================
;
; Copyright (C) 2017-2021 Intel Corporation
;
; SPDX-License-Identifier: MIT
;
;============================ end_copyright_notice =============================

;
; !!! DO NOT EDIT THIS FILE !!!
;
; This file was automagically crafted by GED's model parser.
;

LIBRARY ged
EXPORTS
  GED_GetModelVersionString
  GED_GetReturnValueString
  GED_GetReturnValuePad
  GED_InitEmptyIns
  GED_DecodeIns
  GED_EncodeIns
  GED_InsSize
  GED_InsModified
  GED_CountCompactEncodings
  GED_GetCompactEncodings
  GED_PrintFieldBitLocation
  GED_QueryFieldBitLocation
  GED_FieldSize
  GED_GetOpcode
  GED_GetMnemonic
  GED_SetOpcode
  GED_IsCompact
  GED_GetModel
  GED_GetNumOfSourceOperands
  GED_GetHasDestinationOperand
  GED_GetAccessMode
  GED_SetAccessMode
  GED_GetMaskCtrl
  GED_SetMaskCtrl
  GED_GetDepCtrl
  GED_SetDepCtrl
  GED_GetExecMaskOffsetCtrl
  GED_SetExecMaskOffsetCtrl
  GED_GetChannelOffset
  GED_SetChannelOffset
  GED_GetThreadCtrl
  GED_SetThreadCtrl
  GED_GetPredCtrl
  GED_SetPredCtrl
  GED_GetPredInv
  GED_SetPredInv
  GED_GetExecSize
  GED_SetExecSize
  GED_GetCondModifier
  GED_SetCondModifier
  GED_GetAccWrCtrl
  GED_SetAccWrCtrl
  GED_GetDebugCtrl
  GED_SetDebugCtrl
  GED_GetSaturate
  GED_SetSaturate
  GED_GetDstRegFile
  GED_SetDstRegFile
  GED_GetDstDataType
  GED_SetDstDataType
  GED_GetSrc0RegFile
  GED_SetSrc0RegFile
  GED_GetSrc0DataType
  GED_SetSrc0DataType
  GED_GetSrc1RegFile
  GED_SetSrc1RegFile
  GED_GetSrc1DataType
  GED_SetSrc1DataType
  GED_GetDstChanEn
  GED_SetDstChanEn
  GED_GetDstSubRegNum
  GED_SetDstSubRegNum
  GED_GetDstAddrImm
  GED_SetDstAddrImm
  GED_GetDstRegNum
  GED_SetDstRegNum
  GED_GetDstAddrSubRegNum
  GED_SetDstAddrSubRegNum
  GED_GetDstHorzStride
  GED_SetDstHorzStride
  GED_GetDstAddrMode
  GED_SetDstAddrMode
  GED_GetSrc0ChanSel
  GED_SetSrc0ChanSel
  GED_GetSrc0SubRegNum
  GED_SetSrc0SubRegNum
  GED_GetSrc0AddrImm
  GED_SetSrc0AddrImm
  GED_GetSrc0RegNum
  GED_SetSrc0RegNum
  GED_GetSrc0AddrSubRegNum
  GED_SetSrc0AddrSubRegNum
  GED_GetSrc0SrcMod
  GED_SetSrc0SrcMod
  GED_GetSrc0AddrMode
  GED_SetSrc0AddrMode
  GED_GetSrc0HorzStride
  GED_SetSrc0HorzStride
  GED_GetSrc0Width
  GED_SetSrc0Width
  GED_GetSrc0VertStride
  GED_SetSrc0VertStride
  GED_GetFlagSubRegNum
  GED_SetFlagSubRegNum
  GED_GetFlagRegNum
  GED_SetFlagRegNum
  GED_GetSrc1ChanSel
  GED_SetSrc1ChanSel
  GED_GetSrc1SubRegNum
  GED_SetSrc1SubRegNum
  GED_GetSrc1AddrImm
  GED_SetSrc1AddrImm
  GED_GetSrc1RegNum
  GED_SetSrc1RegNum
  GED_GetSrc1AddrSubRegNum
  GED_SetSrc1AddrSubRegNum
  GED_GetSrc1SrcMod
  GED_SetSrc1SrcMod
  GED_GetSrc1AddrMode
  GED_SetSrc1AddrMode
  GED_GetSrc1HorzStride
  GED_SetSrc1HorzStride
  GED_GetSrc1Width
  GED_SetSrc1Width
  GED_GetSrc1VertStride
  GED_SetSrc1VertStride
  GED_GetImm
  GED_SetImm
  GED_GetSrc2SrcMod
  GED_SetSrc2SrcMod
  GED_GetSrcDataType
  GED_SetSrcDataType
  GED_GetSrc0RepCtrl
  GED_SetSrc0RepCtrl
  GED_GetSrc1RepCtrl
  GED_SetSrc1RepCtrl
  GED_GetSrc2RepCtrl
  GED_SetSrc2RepCtrl
  GED_GetSrc2ChanSel
  GED_SetSrc2ChanSel
  GED_GetSrc2SubRegNum
  GED_SetSrc2SubRegNum
  GED_GetSrc2RegNum
  GED_SetSrc2RegNum
  GED_GetSrc2RegFile
  GED_SetSrc2RegFile
  GED_GetSrc2AddrMode
  GED_SetSrc2AddrMode
  GED_GetSrc2VertStride
  GED_SetSrc2VertStride
  GED_GetSFID
  GED_SetSFID
  GED_GetDescRegFile
  GED_SetDescRegFile
  GED_GetDescDataType
  GED_SetDescDataType
  GED_GetDescAddrSubRegNum
  GED_SetDescAddrSubRegNum
  GED_GetDescRegNum
  GED_SetDescRegNum
  GED_GetDescHorzStride
  GED_SetDescHorzStride
  GED_GetDescWidth
  GED_SetDescWidth
  GED_GetDescVertStride
  GED_SetDescVertStride
  GED_GetMsgDesc
  GED_SetMsgDesc
  GED_GetExMsgDesc
  GED_SetExMsgDesc
  GED_GetEOT
  GED_SetEOT
  GED_GetMathFC
  GED_SetMathFC
  GED_GetJIP
  GED_SetJIP
  GED_GetUIP
  GED_SetUIP
  GED_GetControlIndex
  GED_SetControlIndex
  GED_GetDataTypeIndex
  GED_SetDataTypeIndex
  GED_GetSubRegIndex
  GED_SetSubRegIndex
  GED_GetSrc0Index
  GED_SetSrc0Index
  GED_GetSrc1Index
  GED_SetSrc1Index
  GED_GetDescIndex
  GED_SetDescIndex
  GED_GetExDescRegFile
  GED_SetExDescRegFile
  GED_GetDstMathMacroExt
  GED_SetDstMathMacroExt
  GED_GetSrc0MathMacroExt
  GED_SetSrc0MathMacroExt
  GED_GetSrc1MathMacroExt
  GED_SetSrc1MathMacroExt
  GED_GetSrc2MathMacroExt
  GED_SetSrc2MathMacroExt
  GED_GetBranchCtrl
  GED_SetBranchCtrl
  GED_GetSourceIndex
  GED_SetSourceIndex
  GED_GetSrc2DataType
  GED_SetSrc2DataType
  GED_GetNoSrcDepSet
  GED_SetNoSrcDepSet
  GED_GetExFuncCtrl
  GED_SetExFuncCtrl
  GED_GetExMsgLength
  GED_SetExMsgLength
  GED_GetExDescAddrSubRegNum
  GED_SetExDescAddrSubRegNum
  GED_GetExDescRegNum
  GED_SetExDescRegNum
  GED_GetMsgDescCategory
  GED_SetMsgDescCategory
  GED_GetMsgDescScratchAddrOffset
  GED_SetMsgDescScratchAddrOffset
  GED_GetMsgDescScratchBlockSize
  GED_SetMsgDescScratchBlockSize
  GED_GetMsgDescScratchInvalidateAfterRead
  GED_SetMsgDescScratchInvalidateAfterRead
  GED_GetMsgDescScratchChannelMode
  GED_SetMsgDescScratchChannelMode
  GED_GetMsgDescScratchMessageType
  GED_SetMsgDescScratchMessageType
  GED_GetExecutionDataType
  GED_SetExecutionDataType
  GED_GetSrc0TernaryImm
  GED_SetSrc0TernaryImm
  GED_GetSrc2TernaryImm
  GED_SetSrc2TernaryImm
  GED_GetSrc2HorzStride
  GED_SetSrc2HorzStride
  GED_GetSWSB
  GED_SetSWSB
  GED_GetSrc1IsImm
  GED_SetSrc1IsImm
  GED_GetSrc0IsImm
  GED_SetSrc0IsImm
  GED_GetSrc0SubRegNumByte
  GED_SetSrc0SubRegNumByte
  GED_GetSyncFC
  GED_SetSyncFC
  GED_GetFusionCtrl
  GED_SetFusionCtrl
  GED_GetDataTypeIndexNoDep
  GED_SetDataTypeIndexNoDep
  GED_GetCompactedImm
  GED_SetCompactedImm
  GED_GetRepeatCount
  GED_SetRepeatCount
  GED_GetSystolicDepth
  GED_SetSystolicDepth
  GED_GetSrc2Precision
  GED_SetSrc2Precision
  GED_GetSrc2SubBytePrecision
  GED_SetSrc2SubBytePrecision
  GED_GetSrc1Precision
  GED_SetSrc1Precision
  GED_GetSrc1SubBytePrecision
  GED_SetSrc1SubBytePrecision
  GED_GetBfnFC
  GED_SetBfnFC
  GED_GetExBSO
  GED_SetExBSO
  GED_GetCPS
  GED_SetCPS
  GED_GetSrc1Length
  GED_SetSrc1Length
  GED_GetSrc2IsImm
  GED_SetSrc2IsImm
  GED_GetIndexedSrcAddrImm
  GED_SetIndexedSrcAddrImm
  GED_GetIndexedSrcAddrMode
  GED_SetIndexedSrcAddrMode
  GED_GetIndexedSrcAddrSubRegNum
  GED_SetIndexedSrcAddrSubRegNum
  GED_GetIndexedSrcChanSel
  GED_SetIndexedSrcChanSel
  GED_GetIndexedSrcDataType
  GED_SetIndexedSrcDataType
  GED_GetIndexedSrcHorzStride
  GED_SetIndexedSrcHorzStride
  GED_GetIndexedSrcIndex
  GED_SetIndexedSrcIndex
  GED_GetIndexedSrcIsImm
  GED_SetIndexedSrcIsImm
  GED_GetIndexedSrcMathMacroExt
  GED_SetIndexedSrcMathMacroExt
  GED_GetIndexedSrcRegFile
  GED_SetIndexedSrcRegFile
  GED_GetIndexedSrcRegNum
  GED_SetIndexedSrcRegNum
  GED_GetIndexedSrcRepCtrl
  GED_SetIndexedSrcRepCtrl
  GED_GetIndexedSrcSrcMod
  GED_SetIndexedSrcSrcMod
  GED_GetIndexedSrcSubRegNum
  GED_SetIndexedSrcSubRegNum
  GED_GetIndexedSrcSubRegNumByte
  GED_SetIndexedSrcSubRegNumByte
  GED_GetIndexedSrcVertStride
  GED_SetIndexedSrcVertStride
  GED_GetIndexedSrcWidth
  GED_SetIndexedSrcWidth
  GED_GetArchReg
  GED_SetArchReg
  GED_GetArchRegNum
  GED_SetArchRegNum
  GED_GetSwizzleX
  GED_SetSwizzleX
  GED_GetSwizzleY
  GED_SetSwizzleY
  GED_GetSwizzleZ
  GED_SetSwizzleZ
  GED_GetSwizzleW
  GED_SetSwizzleW
  GED_GetMessageLength
  GED_SetMessageLength
  GED_GetResponseLength
  GED_SetResponseLength
  GED_GetHeaderPresent
  GED_SetHeaderPresent
  GED_GetMessageTypeDP_SAMPLER
  GED_SetMessageTypeDP_SAMPLER
  GED_GetMessageTypeDP_RC
  GED_SetMessageTypeDP_RC
  GED_GetMessageTypeDP_CC
  GED_SetMessageTypeDP_CC
  GED_GetMessageTypeDP_DC0
  GED_SetMessageTypeDP_DC0
  GED_GetTypedSurfaceSlotGroup
  GED_SetTypedSurfaceSlotGroup
  GED_GetTypedAtomicSlotGroup
  GED_SetTypedAtomicSlotGroup
  GED_GetUntypedSurfaceSIMDMode
  GED_SetUntypedSurfaceSIMDMode
  GED_GetUntypedAtomicSIMDMode
  GED_SetUntypedAtomicSIMDMode
  GED_GetInvalidateAfterRead
  GED_SetInvalidateAfterRead
  GED_GetBlockSize
  GED_SetBlockSize
  GED_GetRedChannel
  GED_SetRedChannel
  GED_GetGreenChannel
  GED_SetGreenChannel
  GED_GetBlueChannel
  GED_SetBlueChannel
  GED_GetAlphaChannel
  GED_SetAlphaChannel
  GED_GetReturnDataControl
  GED_SetReturnDataControl
  GED_GetAtomicOperationType
  GED_SetAtomicOperationType
  GED_GetAtomicCounterOperationType
  GED_SetAtomicCounterOperationType
  GED_GetSubFuncID
  GED_SetSubFuncID
  GED_GetBindingTableIndex
  GED_SetBindingTableIndex
  GED_GetFuncControl
  GED_SetFuncControl
  GED_GetMessageTypeDP_DC1
  GED_SetMessageTypeDP_DC1
  GED_GetMessageTypeDP0Category
  GED_SetMessageTypeDP0Category
  GED_GetMessageTypeDP_DC0Legacy
  GED_SetMessageTypeDP_DC0Legacy
  GED_GetMessageTypeDP_DC0ScratchBlock
  GED_SetMessageTypeDP_DC0ScratchBlock
  GED_GetMessageTypeDP_DC2
  GED_SetMessageTypeDP_DC2
  GED_GetMessageTypeDP_DCRO
  GED_SetMessageTypeDP_DCRO
  GED_GetExMessageLength
  GED_SetExMessageLength
  GED_GetDPOpcode
  GED_SetDPOpcode
  GED_GetDPAddrType
  GED_SetDPAddrType
  GED_GetDPVectSize
  GED_SetDPVectSize
  GED_GetDPFlushType
  GED_SetDPFlushType
  GED_GetDPTranspose
  GED_SetDPTranspose
  GED_GetDPFlushRange
  GED_SetDPFlushRange
  GED_GetDPDataSize
  GED_SetDPDataSize
  GED_GetDPFenceScope
  GED_SetDPFenceScope
  GED_GetDPAddrSize
  GED_SetDPAddrSize
  GED_SetRawBits
  GED_GetAccessModeString
  GED_GetAccWrCtrlString
  GED_GetAddrModeString
  GED_GetArchRegString
  GED_GetAtomicOperationTypeString
  GED_GetBlockSizeString
  GED_GetBranchCtrlString
  GED_GetChannelMaskString
  GED_GetChannelModeString
  GED_GetChannelOffsetString
  GED_GetCondModifierString
  GED_GetDataTypeString
  GED_GetDebugCtrlString
  GED_GetDepCtrlString
  GED_GetDpAddrSizeString
  GED_GetDpAddrTypeString
  GED_GetDpDataSizeString
  GED_GetDpFenceScopeString
  GED_GetDpFlushTypeString
  GED_GetDpOpcodeString
  GED_GetDpTransposeString
  GED_GetDpVectSizeString
  GED_GetDstChanEnString
  GED_GetEOTString
  GED_GetExecMaskOffsetCtrlString
  GED_GetExecutionDataTypeString
  GED_GetFusionCtrlString
  GED_GetHeaderPresentString
  GED_GetMaskCtrlString
  GED_GetMathFCString
  GED_GetMathMacroExtString
  GED_GetMessageTypeString
  GED_GetNoSrcDepSetString
  GED_GetOpcodeString
  GED_GetPrecisionString
  GED_GetPredCtrlString
  GED_GetPredInvString
  GED_GetRegFileString
  GED_GetRepCtrlString
  GED_GetReturnDataControlString
  GED_GetSaturateString
  GED_GetSFIDString
  GED_GetSIMDModeString
  GED_GetSlotGroupString
  GED_GetSrcModString
  GED_GetSubBytePrecisionString
  GED_GetSubFuncIDString
  GED_GetSwizzleString
  GED_GetSyncFCString
  GED_GetThreadCtrlString
  GED_GetOperandWidth
  GED_GetOperandNumericType
  GED_GetNibCtrl
