/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef IGA_BXML_OPS_HPP
#define IGA_BXML_OPS_HPP

// ******************** DO NOT MODIFY DIRECTLY ********************
// Auto-generated by IGA project tools

namespace iga
{
    enum class Op
    {
        INVALID = 0,
        FIRST_OP = 1,

        ADD          = FIRST_OP +   0,
        ADD3         = FIRST_OP +   1,
        ADDC         = FIRST_OP +   2,
        AND          = FIRST_OP +   3,
        ASR          = FIRST_OP +   4,
        AVG          = FIRST_OP +   5,
        BFE          = FIRST_OP +   6,
        BFI1         = FIRST_OP +   7,
        BFI2         = FIRST_OP +   8,
        BFN          = FIRST_OP +   9,
        BFREV        = FIRST_OP +  10,
        BRC          = FIRST_OP +  11,
        BRD          = FIRST_OP +  12,
        BREAK        = FIRST_OP +  13,
        CALL         = FIRST_OP +  14,
        CALLA        = FIRST_OP +  15,
        CBIT         = FIRST_OP +  16,
        CMP          = FIRST_OP +  17,
        CMPN         = FIRST_OP +  18,
        CONT         = FIRST_OP +  19,
        CSEL         = FIRST_OP +  20,
        DIM          = FIRST_OP +  21,
        DP2          = FIRST_OP +  22,
        DP3          = FIRST_OP +  23,
        DP4          = FIRST_OP +  24,
        DP4A         = FIRST_OP +  25,
        DPAS         = FIRST_OP +  26,
        DPASW        = FIRST_OP +  27,
        DPH          = FIRST_OP +  28,
        ELSE         = FIRST_OP +  29,
        ENDIF        = FIRST_OP +  30,
        F16TO32      = FIRST_OP +  31,
        F32TO16      = FIRST_OP +  32,
        FBH          = FIRST_OP +  33,
        FBL          = FIRST_OP +  34,
        FRC          = FIRST_OP +  35,
        GOTO         = FIRST_OP +  38,
        HALT         = FIRST_OP +  39,
        IF           = FIRST_OP +  40,
        ILLEGAL      = FIRST_OP +  41,
        JMPI         = FIRST_OP +  42,
        JOIN         = FIRST_OP +  43,
        LINE         = FIRST_OP +  44,
        LRP          = FIRST_OP +  45,
        LZD          = FIRST_OP +  46,
        MAC          = FIRST_OP +  47,
        MACH         = FIRST_OP +  48,
        MACL         = FIRST_OP +  49,
        MAD          = FIRST_OP +  50,
        MADM         = FIRST_OP +  51,
        MATH         = FIRST_OP +  54,
        MOV          = FIRST_OP +  55,
        MOVI         = FIRST_OP +  56,
        MUL          = FIRST_OP +  57,
        NOP          = FIRST_OP +  58,
        NOT          = FIRST_OP +  59,
        OR           = FIRST_OP +  60,
        PLN          = FIRST_OP +  61,
        RET          = FIRST_OP +  62,
        RNDD         = FIRST_OP +  63,
        RNDE         = FIRST_OP +  64,
        RNDU         = FIRST_OP +  65,
        RNDZ         = FIRST_OP +  66,
        ROL          = FIRST_OP +  67,
        ROR          = FIRST_OP +  68,
        SAD2         = FIRST_OP +  69,
        SADA2        = FIRST_OP +  70,
        SEL          = FIRST_OP +  73,
        SEND         = FIRST_OP +  74,
        SENDC        = FIRST_OP +  75,
        SENDS        = FIRST_OP +  76,
        SENDSC       = FIRST_OP +  77,
        SHL          = FIRST_OP +  79,
        SHR          = FIRST_OP +  80,
        SMOV         = FIRST_OP +  81,
        SRND         = FIRST_OP +  82,
        SUBB         = FIRST_OP +  83,
        SYNC         = FIRST_OP +  84,
        WAIT         = FIRST_OP +  85,
        WHILE        = FIRST_OP +  86,
        XOR          = FIRST_OP +  87,

        LAST_OP   = FIRST_OP + 87,
        TOTAL_OPS = LAST_OP
    };
}
#endif // IGA_BXML_OPS_HPP
