/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

/*
 * !!! DO NOT EDIT THIS FILE !!!
 *
 * This file was automagically crafted by GED's model parser.
 */


#ifndef GED_ENCODING_MASKS_TABLES_H
#define GED_ENCODING_MASKS_TABLES_H

#include "common/ged_ins_encoding_masks.h"
extern const ged_instruction_masks_entry_t EncodingMasksTable0[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable1[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable2[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable3[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable4[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable5[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable6[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable7[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable8[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable9[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable10[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable11[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable12[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable13[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable14[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable15[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable16[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable17[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable18[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable19[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable20[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable21[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable22[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable23[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable24[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable25[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable26[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable27[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable28[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable29[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable30[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable31[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable32[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable33[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable34[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable35[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable36[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable37[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable38[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable39[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable40[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable41[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable42[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable43[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable44[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable45[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable46[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable47[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable48[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable49[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable50[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable51[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable52[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable53[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable54[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable55[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable56[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable57[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable58[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable59[12];
extern const ged_instruction_masks_entry_t EncodingMasksTable60[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable61[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable62[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable63[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable64[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable65[10];
extern const ged_instruction_masks_entry_t EncodingMasksTable66[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable67[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable68[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable69[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable70[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable71[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable72[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable73[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable74[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable75[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable76[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable77[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable78[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable79[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable80[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable81[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable82[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable83[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable84[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable85[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable86[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable87[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable88[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable89[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable90[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable91[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable92[12];
extern const ged_instruction_masks_entry_t EncodingMasksTable93[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable94[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable95[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable96[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable97[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable98[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable99[8];
extern const ged_instruction_masks_entry_t EncodingMasksTable100[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable101[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable102[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable103[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable104[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable105[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable106[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable107[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable108[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable109[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable110[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable111[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable112[7];
extern const ged_instruction_masks_entry_t EncodingMasksTable113[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable114[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable115[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable116[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable117[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable118[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable119[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable120[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable121[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable122[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable123[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable124[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable125[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable126[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable127[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable128[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable129[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable130[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable131[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable132[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable133[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable134[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable135[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable136[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable137[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable138[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable139[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable140[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable141[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable142[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable143[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable144[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable145[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable146[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable147[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable148[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable149[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable150[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable151[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable152[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable153[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable154[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable155[128];
extern const ged_instruction_masks_entry_t EncodingMasksTable156[9];
extern const ged_instruction_masks_entry_t EncodingMasksTable157[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable158[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable159[5];
extern const ged_instruction_masks_entry_t EncodingMasksTable160[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable161[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable162[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable163[16];
extern const ged_instruction_masks_entry_t EncodingMasksTable164[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable165[6];
extern const ged_instruction_masks_entry_t EncodingMasksTable166[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable167[3];
extern const ged_instruction_masks_entry_t EncodingMasksTable168[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable169[4];
extern const ged_instruction_masks_entry_t EncodingMasksTable170[2];
extern const ged_instruction_masks_entry_t EncodingMasksTable171[3];
#endif // GED_ENCODING_MASKS_TABLES_H
