/*========================== begin_copyright_notice ============================

Copyright (C) 2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

// Platform Revision Ids for ACM

// AUTO-GENERATED FILE. DO NOT EDIT.

#include "wa_def.h"

#define ACM_G10_GT_REV_ID_A0   SI_REV_ID(0,0)
#define ACM_G10_GT_REV_ID_A1   SI_REV_ID(1,1)
#define ACM_G10_GT_REV_ID_B0   SI_REV_ID(4,4)
#define ACM_G10_GT_REV_ID_C0   SI_REV_ID(8,8)
#define ACM_G10_GT_REV_ID_C1   SI_REV_ID(9,9)

