/*========================== begin_copyright_notice ============================

Copyright (C) 2020-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef IGCLLVM_PASS_TIMING_INFO_H
#define IGCLLVM_PASS_TIMING_INFO_H

#include "llvm/Config/llvm-config.h"
#if LLVM_VERSION_MAJOR > 7
#include "llvm/IR/PassTimingInfo.h"
#endif

#endif
