/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#pragma once

//
// if GTGPU is defined, the definitions of the two routines are
// in GTGPU runtime
//
extern "C" void* allocCodeBlock(size_t sz);
