/*========================== begin_copyright_notice ============================

Copyright (C) 2018-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef IGCLLVM_TRANSFORMS_UTILS_INSTSIMPLIFYPASS_H
#define IGCLLVM_TRANSFORMS_UTILS_INSTSIMPLIFYPASS_H

#include "llvm/Config/llvm-config.h"

#if LLVM_VERSION_MAJOR >= 7
#include "llvm/Transforms/IPO.h"
#endif

#endif
