/*========================== begin_copyright_notice ============================

Copyright (C) 2017-2021 Intel Corporation

SPDX-License-Identifier: MIT

============================= end_copyright_notice ===========================*/

#ifndef __IFCVT_H__
#define __IFCVT_H__

#include "FlowGraph.h"

void runIfCvt(vISA::FlowGraph &);

#endif // __IFCVT_H__

// vim:ts=4:sw=4:et:
